#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <vector>
#include <string>
#include <map>

//struktura vzorku
struct Sample {
	std::vector<double> features;	//priznaky vzorku
	size_t label;					//klasifikacni trida
};

bool read_input_csv_type0(const std::string& infile, std::vector<Sample>& target) {

	std::ifstream ifs(infile);

	try {
		std::string line;
		while (std::getline(ifs, line)) {

			std::istringstream iss(line);

			std::string field;
			if (!std::getline(iss, field, ',')) {
				return false;
			}

			size_t label = std::stoull(field);

			target.push_back({ {}, label });

			std::vector<double>& features = target.rbegin()->features;

			while (std::getline(iss, field, ',')) {
				 features.push_back(std::stod(field));
			}

		}

	}
	catch (...) {
		return false;
	}

	return true;
}

bool read_input_csv_type1(const std::string& infile, std::vector<Sample>& target) {

	std::ifstream ifs(infile);

	try {
		std::string line;
		while (std::getline(ifs, line)) {

			std::istringstream iss(line);

			std::string field;
			if (!std::getline(iss, field, ',')) {
				return false;
			}

			size_t label = std::stoull(field);

			//zahodit
			if (!std::getline(iss, field, ',')) {
				return false;
			}

			target.push_back({ {}, label });

			std::vector<double>& features = target.rbegin()->features;

			while (std::getline(iss, field, ',')) {
				features.push_back(std::stod(field));
			}

		}

	}
	catch (...) {
		return false;
	}

	return true;
}

bool read_input_csv_type3(const std::string& infile, std::vector<Sample>& target) {

	std::ifstream ifs(infile);

	try {
		std::string line;
		while (std::getline(ifs, line)) {

			std::istringstream iss(line);
			std::string field;

			target.push_back({ {}, 0 });

			std::vector<double>& features = target.rbegin()->features;
			size_t lastRead = 0;

			while (std::getline(iss, field, ' ')) {
				features.push_back(std::stod(field));
				try
				{
					lastRead = std::stoull(field);
				}
				catch (...) {}
			}

			features.erase(features.begin() + (features.size() - 1));
			target.rbegin()->label = lastRead;

		}

	}
	catch (...) {
		return false;
	}

	return true;
}

bool read_input_csv_type4(const std::string& infile, std::vector<Sample>& target) {

	std::ifstream ifs(infile);

	try {
		std::string line;
		while (std::getline(ifs, line)) {

			std::istringstream iss(line);
			std::string field;

			target.push_back({ {}, 0 });

			std::vector<double>& features = target.rbegin()->features;
			size_t lastRead = 0;

			while (std::getline(iss, field, ',')) {
				features.push_back(std::stod(field));
				try
				{
					lastRead = std::stoull(field);
				}
				catch (...) {}
			}

			features.erase(features.begin() + (features.size() - 1));
			target.rbegin()->label = lastRead;

		}

	}
	catch (...) {
		return false;
	}

	return true;
}

bool read_input_csv_type5(const std::string& infile, std::vector<Sample>& target) {

	std::ifstream ifs(infile);

	std::map<std::string, size_t> labelsStr;
	size_t nextLabel = 0;

	try {
		std::string line;
		while (std::getline(ifs, line)) {

			std::istringstream iss(line);
			std::string field;

			target.push_back({ {}, 0 });

			std::vector<double>& features = target.rbegin()->features;

			while (std::getline(iss, field, ';')) {
				try {
					features.push_back(std::stod(field));
				}
				catch (...)
				{
					break;
				}
			}

			if (labelsStr.find(field) == labelsStr.end()) {
				labelsStr[field] = nextLabel++;
			}

			target.rbegin()->label = labelsStr[field];
		}

	}
	catch (...) {
		return false;
	}

	return true;
}

bool read_input_internal(const std::string& infile, std::vector<Sample>& target, size_t featureCount) {

	std::ifstream ifs(infile);

	try {
		std::string line;

		//zahodit pocet radek
		std::getline(ifs, line);

		while (std::getline(ifs, line)) {

			std::istringstream iss(line);
			std::string field;

			target.push_back({ {}, 0 });

			std::vector<double>& features = target.rbegin()->features;

			size_t featureIdx = 0;
			while (std::getline(iss, field, ' ') && featureIdx <= featureCount) {
				features.push_back(std::stod(field));
				featureIdx++;
			}

			size_t curClass = 0;
			while (std::getline(iss, field, ' ')) {
				if (field == "1") {
					break;
				}
				curClass++;
			}

			target.rbegin()->label = curClass;
		}

	}
	catch (std::exception& ex) {
		std::cerr << ex.what() << std::endl;
		return false;
	}

	return true;
}

void subtractMean(std::vector<Sample>& target) {

	const size_t featureCount = target[0].features.size();

	std::vector<double> means(featureCount);
	std::vector<double> stdevs(featureCount);
	std::fill(means.begin(), means.end(), 0);
	std::fill(stdevs.begin(), stdevs.end(), 0);

	for (auto& t : target) {
		for (size_t i = 0; i < featureCount; i++) {
			means[i] += t.features[i];
		}
	}

	for (size_t i = 0; i < featureCount; i++) {
		means[i] /= static_cast<double>(target.size());
	}

	for (auto& t : target) {
		for (size_t i = 0; i < featureCount; i++) {
			stdevs[i] += std::pow(t.features[i] - means[i], 2);
		}
	}

	for (size_t i = 0; i < featureCount; i++) {
		stdevs[i] = std::sqrt(stdevs[i] / static_cast<double>(target.size()));
	}

	for (auto& t : target) {
		for (size_t i = 0; i < featureCount; i++) {
			t.features[i] = (t.features[i] - means[i]) / (stdevs[i] + 0.0000001);
		}
	}

}

int main(int argc, char** argv) {

	size_t featureCount = 0;
	std::string operation = "";

	if (argc < 4) {
		std::cerr << "Pouziti: " << argv[0] << " <vstupni soubor> <vystupni soubor> <typ vstupu> [pocet priznaku] [operace]" << std::endl
			<< std::endl
			<< "Typ vstupu:" << std::endl
			<< "    0 = CSV, prvni oznaceni tridy, nasleduje pole priznaku" << std::endl
			<< "    1 = CSV, prvni oznaceni tridy, druhe poradove cislo, nasleduje pole priznaku" << std::endl
			<< "    2 = vystupni format" << std::endl
			<< "    3 = SSV, pole priznaku, posledni cislo oznacuje tridu" << std::endl
			<< "    4 = CSV, pole priznaku, posledni cislo oznacuje tridu" << std::endl
			<< "    5 = CSV se strednikem, pole priznaku, posledni cislo oznacuje tridu reprezentovane retezcem" << std::endl
			<< std::endl
			<< "Operace: " << std::endl
			<< "    mean = odecteni stredni hodnoty vsech prvku vektoru" << std::endl;
		return 1;
	}

	const std::string infile = argv[1];
	const std::string outfile = argv[2];
	size_t intype = 0;
	try {
		intype = std::stoull(argv[3]);
	}
	catch (...) {
		std::cerr << "Neplatny typ" << std::endl;
		return 2;
	}

	if (argc > 4) {
		try
		{
			featureCount = std::stoull(argv[4]);
		}
		catch (...)
		{
			//
		}
	}

	if (argc > 5) {
		operation = argv[5];
	}

	std::vector<Sample> samples;
	bool success = false;

	switch (intype) {
		case 0:
			success = read_input_csv_type0(infile, samples);
			break;
		case 1:
			success = read_input_csv_type1(infile, samples);
			break;
		case 2:
			success = read_input_internal(infile, samples, featureCount);
			break;
		case 3:
			success = read_input_csv_type3(infile, samples);
			break;
		case 4:
			success = read_input_csv_type4(infile, samples);
			break;
		case 5:
			success = read_input_csv_type5(infile, samples);
			break;
	}

	if (!success) {
		std::cerr << "Nepodarilo se nacist vstup" << std::endl;
		return 3;
	}

	if (operation == "mean") {
		subtractMean(samples);
	}

	size_t minLabel = samples[0].label;
	size_t maxLabel = samples[0].label;

	for (const auto& s : samples) {
		if (s.label < minLabel) {
			minLabel = s.label;
		}
		else if (s.label > maxLabel) {
			maxLabel = s.label;
		}
	}

	std::cout << "feature count: " << samples[0].features.size() << std::endl;
	std::cout << "labels: " << minLabel << " - " << maxLabel << std::endl;

	std::ofstream ofs(outfile);

	ofs << samples.size() << '\n';
	for (const auto& s : samples) {
		for (double d : s.features) {
			ofs << d << ' ';
		}

		for (size_t m = minLabel; m <= maxLabel; m++) {
			ofs << (m == s.label) ? '1' : '0';
			if (m != maxLabel) {
				ofs << ' ';
			}
		}

		ofs << '\n';
	}

	return 0;
}
