#include "mlp.h"

#include <chrono>

thread_local std::vector<std::vector<double>> tlDeltaBuffers;
thread_local bool tlDeltaBuffersInit = false;

double backpropagation(std::unique_ptr<NETWORK>& n, const std::vector<std::vector<double>>& sample, size_t startIndex, size_t count, const std::vector<std::vector<double>>& desired, double learningRate, double momentumRate, size_t epochLimit, bool silent, bool allowDiverge) {

	double diff;
	double lastError, curError;
	size_t epoch = 0;

	if (!tlDeltaBuffersInit) {

		tlDeltaBuffersInit = true;
		tlDeltaBuffers.resize(n->layers.size());
		for (size_t j = 0; j < n->layers.size(); j++) {
			tlDeltaBuffers[j].resize(n->layers[j].neurons.size());
		}
	}

	lastError = avgQuadError(n, sample, startIndex, count, desired);
	size_t microsAccumulator = 0;

	do {		//backprop
		if (!silent) {
			std::cout << "Epocha " << epoch << std::endl;
		}

		auto time_start = std::chrono::steady_clock::now();

		for (size_t i = startIndex; i < startIndex + count; i++) {

			const auto& grad_sample = sample[i];
			const auto& grad_desired = desired[i];

			for (size_t j = n->layers.size() - 1; j > 0; j--) {
				getDelta(n, grad_sample, grad_desired, j, tlDeltaBuffers);
				adjustWeights(n, j, tlActivations[j - 1], tlDeltaBuffers[j], learningRate, momentumRate);
			}

			getDelta(n, grad_sample, grad_desired, 0, tlDeltaBuffers);
			adjustWeights(n, 0, grad_sample, tlDeltaBuffers[0], learningRate, momentumRate);
		}

		curError = avgQuadError(n, sample, startIndex, count, desired);

		if (lastError < curError && !allowDiverge) {
			if (!silent) {
				std::cout << "Diverguje!" << std::endl;
			}
			break;
		}

		diff = fabs(lastError - curError);
		lastError = curError;

		auto time_end = std::chrono::steady_clock::now();

		microsAccumulator += std::chrono::duration_cast<std::chrono::microseconds>(time_end - time_start).count();

		epoch++;

		if (!silent) {
			std::cout << "Odchylka " << lastError << std::endl;
			gPopulationStats.push_back({ epoch, microsAccumulator, lastError, lastError });
		}

	} while (diff > EPSILON && (epochLimit == 0 || epoch < epochLimit));

	return lastError;
}
