#include "mlp.h"
#include "genetic_algorithm.h"

#include <thread>
#include <iomanip>
#include <sstream>
#include <algorithm>
#include <random>

//mira uceni neuronove site
double RATE = 0.1;
//setrvacnost pri uceni zpetnym sirenim
double MOMENTUM = 0.2;

//zpusob uceni, pokud neni zadan je to backprop
LearnType learnType = LearnType::Backpropagation;

//pocet vah co uz bylo nacteno
int initW = 0;
//nactene vahy ze souboru
std::vector<double> wValues;
//mod metody blitz
int methodBlitz = 0;
//mod metody freely growing
int methodFreely = 0;
//velikost populace genetickeho algoritmu
size_t population_size_request = 10;
//pocet generaci genetickeho algoritmu
size_t generations_count_request = 1000;
//rezim validace
bool validationMode = false;
//rezim klasifikace
bool classificationMode = false;
//nahodne zamichani vzorku
bool shuffle = false;

//nazev vstupniho souboru
std::string inputFile = "input.txt";
//soubor s modelem
std::string modelFile = "model.txt";

//statistiky konvergence
std::vector<PopulationStatistics> gPopulationStats;

//
static std::random_device _rdev;
static std::default_random_engine _reng;
static std::uniform_real_distribution<double> _rnorm_dist(-1.0, 1.0);
thread_local std::vector<std::vector<double>> tlPreactivations;
thread_local std::vector<std::vector<double>> tlActivations;
thread_local bool tlActBufferInit = false;

//klonovani neuronove site
std::unique_ptr<NETWORK> cloneNetwork(std::unique_ptr<NETWORK>& n, bool randWeights) {

	std::unique_ptr<NETWORK> cloned = std::make_unique<NETWORK>();

	cloned->inputs = n->inputs;
	cloned->layers.resize(n->layers.size());

	for (size_t l = 0; l < n->layers.size(); l++) {
		cloned->layers[l].neurons.resize(n->layers[l].neurons.size());

		for (size_t u = 0; u < n->layers[l].neurons.size(); u++) {

			if (randWeights) {

				cloned->layers[l].neurons[u].weights.resize(n->layers[l].neurons[u].weights.size());
				cloned->layers[l].neurons[u].last_weights.resize(n->layers[l].neurons[u].weights.size());

				for (size_t w = 0; w < cloned->layers[l].neurons[u].weights.size(); w++) {
					cloned->layers[l].neurons[u].weights[w] = drand();
					cloned->layers[l].neurons[u].last_weights[w] = cloned->layers[l].neurons[u].weights[w];
				}
			}
			else {
				cloned->layers[l].neurons[u].weights = n->layers[l].neurons[u].weights;
				cloned->layers[l].neurons[u].last_weights = n->layers[l].neurons[u].last_weights;
			}
		}
	}

	return cloned;
}

//kopirovani vah jedne site do druhe
void cloneNetworkWeights(std::unique_ptr<NETWORK>& n, std::unique_ptr<NETWORK>& target) {

	for (size_t l = 0; l < n->layers.size(); l++) {

		for (size_t u = 0; u < n->layers[l].neurons.size(); u++) {

			target->layers[l].neurons[u].weights = n->layers[l].neurons[u].weights;
			target->layers[l].neurons[u].last_weights = n->layers[l].neurons[u].last_weights;
		}
	}
}

int main(int argc, char** argv) {
	int inputs;				//pocet neuronu vstupni vrstvy
	int outputs;			//pocet neuronu vystupni vrstvy
	int samples = -1;		//pocet vzorku
	int epoch = 0;
	int layers;					//pocet vrstev bez vstupni
	std::vector<int> npl;		//neurons per layer
	int showEpochs = 0;
	int notShowWeights = 0;
	int showEqms = 0;
	double diff;

	checkArgs(argc, argv, inputs, outputs, samples, layers, npl, initW, wValues, showEqms, showEpochs, notShowWeights);

	if (argc <= 1) {		//spatne vstupy -> vypis
		showUsage();
		return 1;
	}

	if (initW > 0) {		//nacte vahy do wValues
		std::ifstream weights(modelFile);
		std::string line;

		if (!weights.is_open() || !weights) {
			std::cerr << "Nelze otevrit soubor s modelem" << std::endl;
			return 1;
		}

		for (int i = 0; i < initW && weights; i++) {
			weights >> wValues[i];
		}

		weights.close();
	}

	std::unique_ptr<NETWORK> n = createNetwork(inputs, layers, npl, (initW > 0));		//vytvori neuronovou sit

	if (!classificationMode && !validationMode) {		//trenovaci rezim

		std::ifstream inputfile(inputFile);

		if (!inputfile.is_open() || !inputfile) {
			std::cerr << "Nelze otevrit vstupni soubor" << std::endl;
			return 1;
		}

		std::cout << "Nacitani vstupu" << std::endl;

		inputfile >> samples;

		std::vector<std::vector<double>> sample(samples);		//2D vektor vzorku
		std::vector<std::vector<double>> desired(samples);		//2D vektor odpovedi ucitele

		for (int c = 0; c < samples; c++) {		//pres vzorky
			sample[c].resize(inputs + 1);		//prida prahovaci neuron
			desired[c].resize(outputs);

			sample[c][0] = -1.0;

			for (int i = 1; i <= inputs && inputfile; i++) {		//pres neurony vstupu
				inputfile >> sample[c][i];
			}

			for (int i = 0; i < outputs && inputfile; i++) {		//pres zbytek = odpovedi ucitele
				inputfile >> desired[c][i];
			}
		}

		if (shuffle) {		//zamichani vzorku na vstupu

			std::cout << "Nahodne zamichani vzorku" << std::endl;

			const auto seed = _rdev();

			//nastavime seed na pevne cislo
			_reng.seed(seed);
			std::shuffle(sample.begin(), sample.end(), _reng);
			//zopakujeme totez pro pozadovane prvky s novou inicializaci generatoru aby se vygenerovala stejna posloupnost
			_reng.seed(seed);
			std::shuffle(desired.begin(), desired.end(), _reng);
		}

		//znovu inicializujeme nahodna cisla
		_reng.seed(_rdev());

		if (methodBlitz > 0) {	//inicializace site blitz
			std::cout << "Inicializace neuronove site metodou blitz" << std::endl;
			blitzInit(n, sample, desired, methodBlitz);
		}

		if (methodFreely) {		//inicializace site freely growing
			std::cout << "Inicializace neuronove site metodou freely growing" << std::endl;
			freelyGrowingInit(n, sample, desired, methodFreely);
		}

		std::cout << "Uceni" << std::endl;

		diff = avgQuadError(n, sample, 0, sample.size(), desired);
		std::cout << "Prumerna chyba pred ucenim = " << diff << std::endl;

		gPopulationStats.push_back({ 0, 0, diff, diff });

		//podle zvolene metody uceni se nyni zapocne uceni

		if (learnType == LearnType::Backpropagation) {
			diff = backpropagation(n, sample, 0, samples, desired, RATE, MOMENTUM);
		}
		else if (learnType == LearnType::Genetic_Algorithm_Backprop) {
			diff = genetic_algorithm<true, true>(n, sample, samples, desired, population_size_request, generations_count_request);
		}
		else if (learnType == LearnType::Genetic_Algorithm_NoBackprop) {
			diff = genetic_algorithm<false, true>(n, sample, samples, desired, population_size_request, generations_count_request);
		}
		else if (learnType == LearnType::Meta_Backprop) {
			diff = genetic_algorithm<true, false>(n, sample, samples, desired, population_size_request, generations_count_request);
		}

		saveWeights(n, modelFile);	//ulozeni vah do souboru
		savePopulationStats();		//ulozeni statistik konvergence do souboru

		diff = avgQuadError(n, sample, 0, samples, desired);	//spocte se prumerna chyba

		std::cout << "Prumerna chyba po uceni " << diff << std::endl;

	}
	else if (!classificationMode && validationMode) { //rezim validace
		std::vector<double> input(inputs + 1);
		std::vector<double> output(outputs);

		std::ifstream validationData(inputFile);

		size_t validDataCount = 0;
		validationData >> validDataCount;		//nacte se pocet vzorku ze souboru

		size_t success = 0;

		//pro vsechny vzorky se provede klasifikace a vyhodnoceni
		for (size_t smp = 0; smp < validDataCount && validationData; smp++) {

			std::string line;

			input[0] = -1.0;	//prida se bias
			for (int i = 1; i <= inputs && validationData; i++) {
				validationData >> input[i];
			}

			//najde se pozadovana trida
			int desiredClass = -1;
			for (int i = 0; i < outputs; i++) {
				validationData >> output[i];
				if (output[i] > 0.5) {
					desiredClass = i;
				}
			}

			//pokud se povedlo nacist cely vstup tak se provede klasifikace a overeni
			if (validationData) {
				//aktivuje se sit
				calculateNetworkActivations(n, input);
				std::vector<double> out = tlActivations[layers - 1];

				int classified = -1;
				double maxvalue = -1;

				//najde se trida
				for (int i = 0; i < outputs; i++) {
					if (out[i] > maxvalue) {
						classified = i;
						maxvalue = out[i];
					}
				}

				//pokud trida odpovida nactene tak se pricte uspesne klasifikovany prvek
				if (desiredClass == classified) {
					success++;
				}
			}
		}

		const double successRate = static_cast<double>(success) / static_cast<double>(validDataCount);	//spocteni procentualni uspesnosti klasifikace

		std::cout << "Celkem vzorku: " << validDataCount << std::endl;
		std::cout << "Spravne klasifikovano: " << success << " (" << std::setprecision(4) << 100.0*successRate << " %)" << std::endl;
	}
	else if (classificationMode) {				//klasifikace

		std::vector<double> input(inputs + 1);
		std::ifstream klasifData(inputFile);

		while (klasifData) {	//dokud je co cist ze souboru

			std::string line;

			input[0] = -1.0;	//prida se bias a nactou se data
			for (int i = 1; i <= inputs && klasifData; i++) {
				klasifData >> input[i];
			}

			if (klasifData) {	//pokud se nacetlo vse
				//aktivuje se sit
				calculateNetworkActivations(n, input);
				std::vector<double> out = tlActivations[layers - 1];

				int classified = -1;
				double maxvalue = -1;

				//najde se do jake tridy se klasifikoval vzorek (nejvetsi aktivace vystupniho neuronu)
				for (int i = 0; i < outputs; i++) {
					std::cout << "y" << i << ": " << out[i] << '\t';

					if (out[i] > maxvalue) {
						classified = i;
						maxvalue = out[i];
					}
				}

				std::cout << "class = " << classified;
				std::cout << "\n---" << std::endl;
			}
		}
	}

	return 0;
}

//vytvori model neur. site
std::unique_ptr<NETWORK> createNetwork(size_t inputs, size_t layers, std::vector<int>& layerNeurons, bool fromLoadedWeights) {
	std::unique_ptr<NETWORK> n = std::make_unique<NETWORK>();		//neur.sit

	n->inputs = inputs;				//pocet neuronu vstupni vrstvy
	n->layers.resize(layers);		//pole vrstev inic.

	size_t wCount = 0;

	for (size_t i = 0; i < layers; i++) {			//pro kazdou vrstvu

		int neurons = layerNeurons[i];			//pocet neuronu v akt. vrstve

		LAYER& currLayer = n->layers[i];		//akt. vrstvy
		currLayer.neurons.resize(neurons);		// pole neuronu akt. vrstvy

		for (size_t j = 0; j < neurons; j++) {		//pro kazdy neuron v akt. vrstve

			NEURON& neuron = currLayer.neurons[j];		//akt. neuron
			size_t weights_count = (i == 0 ? n->inputs : layerNeurons[i - 1]) + 1;	//pocet vah vstupujicich do neuronu + prahovaci (theta)
			neuron.weights.resize(weights_count);			//pole vah vstupujicich do neuronu
			neuron.last_weights.resize(weights_count);		

			for (size_t k = 0; k < neuron.weights.size(); k++) {		//pres vahy akt. neuronu
				neuron.weights[k] = fromLoadedWeights ? wValues[wCount++] : drand();				//nastavi vahy na nahodne hodnoty
				neuron.last_weights[k] = neuron.weights[k];				//nastavi vahy predchozi iterace na soucasne
			}
		}

	}

	return n;
}

//ulozi vahy do docasneho souboru kdyby doslo k predcasnemu ukonceni
void saveTempWeights(std::unique_ptr<NETWORK>& n) {
	saveWeights(n, modelFile + ".tmp");
}

//vypise vahy do souboru
void saveWeights(std::unique_ptr<NETWORK>& n, const std::string& outputModelFile) {

	std::ofstream model(outputModelFile);

	const size_t layers = n->layers.size();				//pocet vrstev

	for (size_t i = 0; i < layers; i++) {		//pres vrstvy site
		const LAYER& l = n->layers[i];				//akt. vrstva

		for (size_t j = 0; j < l.neurons.size(); j++) {		//pres neurony vrstvy
			const NEURON& neuron = l.neurons[j];		//akt. neuron

			for (size_t k = 0; k < neuron.weights.size(); k++) {		//pres vahy neuronu
				model << neuron.weights[k] << '\n';
			}
		}
	}
}

//ulozi statistiky konvergence do souboru
void savePopulationStats() {

	std::ofstream popsf("stat." + modelFile + ".csv");

	std::sort(gPopulationStats.begin(), gPopulationStats.end(), [](const PopulationStatistics& p1, const PopulationStatistics& p2) {
		return p1.generation < p2.generation;
	});

	popsf << "generation,real_time,best_fitness,average_fitness" << std::endl;

	for (auto& p : gPopulationStats) {
		popsf << p.generation << "," << p.microseconds << "," << p.bestFitness << "," << p.avgFitness << std::endl;
	}
}

//aktivacni funkce
double sigmoid(double x) {
	return 1.0 / (1.0 + std::exp(-x));			//logisticka sigmoida
}

//derivace aktivacni funkce
double activationDeriv(double x) {
	return sigmoid(x) * (1.0 - sigmoid(x));		//derivace logisticke sigmoidy
}

//generuje nahodne cislo pro vahu neuronu
double drand() {
	return _rnorm_dist(_reng);
}

//vypocte aktivace cele site (vsech vrstev a jejich neuronu)
void calculateNetworkActivations(std::unique_ptr<NETWORK>& n, const std::vector<double>& sample) {

	const size_t lastLayerIdx = n->layers.size() - 1;

	//inicializuje se buffer pokud to jeste nebylo udelano v tomto vlakne
	if (!tlActBufferInit) {
		tlActBufferInit = true;

		tlPreactivations.resize(n->layers.size());
		tlActivations.resize(n->layers.size());

		for (size_t i = 0; i < n->layers.size(); i++) {
			const size_t currLaySize = n->layers[i].neurons.size();			//pocet neuronu v akt. vrstve

			tlPreactivations[i].resize(currLaySize);
			tlActivations[i].resize(currLaySize + ((i == n->layers.size() - 1) ? 0 : 1));
			tlActivations[i][0] = -1.0;
		}
	}

	//aktivace 0. vrstvy
	{
		const LAYER& currLayer = n->layers[0];			//aktualni vrstva (krom vstupni)
		const size_t currLaySize = currLayer.neurons.size();			//pocet neuronu v akt. vrstve

		std::vector<double>& preactivations = tlPreactivations[0];
		std::vector<double>& activations = tlActivations[0];				//pole pro aktivace neuronu

		const size_t wLen = currLayer.neurons[0].weights.size();		//pocet vah pro neurony v akt. vrstve

		for (size_t i = 0; i < currLaySize; i++) {			//pres neurony v akt.vrstve

			const NEURON& currNeuron = currLayer.neurons[i];		//akturalni neuron

			preactivations[i] = currNeuron.weights[0] * sample[0];			//predvysledek akt. neuronu na 0
			for (size_t j = 1; j < wLen; j++) {			//pres vahy akt. neuronu
				preactivations[i] += currNeuron.weights[j] * sample[j];
			}

			activations[i + 1] = sigmoid(preactivations[i]);
		}
	}

	//aktivace dalsich skrytych vrstev
	for (size_t layerIdx = 1; layerIdx < lastLayerIdx; layerIdx++) {

		const LAYER& currLayer = n->layers[layerIdx];		//akt. vrstva
		const size_t currLaySize = currLayer.neurons.size();		//pocet neuronu v akt. vrtsve

		std::vector<double>& preactivations = tlPreactivations[layerIdx];
		std::vector<double>& activations = tlActivations[layerIdx];				//pole pro aktivace neuronu

		const size_t wLen = currLayer.neurons[0].weights.size();		//pocet vah pro neurony v akt. vrstve

		const std::vector<double>& prevActivation = tlActivations[layerIdx - 1];		//aktivace neuronu z predchozi vrstvy

		for (size_t i = 0; i < currLaySize; i++) {			//pres neurony v akt.vrstve

			const NEURON& currNeuron = currLayer.neurons[i];		//akturalni neuron

			preactivations[i] = currNeuron.weights[0] * prevActivation[0];			//predvysledek akt. neuronu na 0
			for (size_t j = 1; j < wLen; j++) {			//pres vahy akt. neuronu
				preactivations[i] += currNeuron.weights[j] * prevActivation[j];
			}

			activations[i + 1] = sigmoid(preactivations[i]);
		}
	}

	//aktivace vystupni vrstvy
	{
		const LAYER& currLayer = n->layers[lastLayerIdx];		//akt. vrstva
		const size_t currLaySize = currLayer.neurons.size();		//pocet neuronu v akt. vrtsve

		std::vector<double>& preactivations = tlPreactivations[lastLayerIdx];
		std::vector<double>& activations = tlActivations[lastLayerIdx];				//pole pro aktivace neuronu

		const size_t wLen = currLayer.neurons[0].weights.size();		//pocet vah pro neurony v akt. vrstve

		const std::vector<double>& prevActivation = tlActivations[lastLayerIdx - 1];		//aktivace neuronu z predchozi vrstvy

		for (size_t i = 0; i < currLaySize; i++) {			//pres neurony v akt.vrstve

			const NEURON& currNeuron = currLayer.neurons[i];		//akturalni neuron

			preactivations[i] = currNeuron.weights[0] * prevActivation[0];			//predvysledek akt. neuronu na 0
			for (size_t j = 1; j < wLen; j++) {			//pres vahy akt. neuronu
				preactivations[i] += currNeuron.weights[j] * prevActivation[j];
			}

			activations[i] = sigmoid(preactivations[i]);		//prozene sigmoidou
		}
	}
}

//kvadraticka chyba klasifikace nad jednim prvkem
double quadraticError(std::unique_ptr<NETWORK>& n, const std::vector<double>& sample, const std::vector<double>& desired) {

	calculateNetworkActivations(n, sample);

	const std::vector<double>& lastLayerActivations = tlActivations[n->layers.size() - 1];		//aktivace vystupni vrstvy
	const size_t outputs = n->layers[n->layers.size() - 1].neurons.size();			//pocet neuronu vystupni vrstvy

	double e = 0.0;			//vysledna chyba
	for (size_t i = 0; i < outputs; i++) {			//pres neurony vystupni vrstvy
		const double le = desired[i] - lastLayerActivations[i];
		e += le * le;		//pricte kvadratickou chyba jednoho neuronu
	}

	return e;//*0.5;
}

//prumerna kvadraticka odchylka
double avgQuadError(std::unique_ptr<NETWORK>& n, const std::vector<std::vector<double>>& samples, size_t startIndex, size_t count, const std::vector<std::vector<double>>& desired) {

	double eqm = 0.0;			//celkova chyba ze vsech vstupu
	for (size_t i = startIndex; i < startIndex + count; i++) {
		eqm += quadraticError(n, samples[i], desired[i]);
	}

	return eqm / static_cast<double>(count);		//prumerna chyba
}

//vypocita gradient v kazde vrstve a neuronu
void getDelta(std::unique_ptr<NETWORK>& n, const std::vector<double>& sample, const std::vector<double>& desired, size_t layer, std::vector<std::vector<double>>& deltabuffers) {

	const size_t lastLayerIdx = n->layers.size() - 1;
	calculateNetworkActivations(n, sample);

	//posledni vrstva pocita s odpovedi ucitele
	{
		const std::vector<double>& preactivations = tlPreactivations[lastLayerIdx];
		const std::vector<double>& activations = tlActivations[lastLayerIdx];
		std::vector<double>& delta = deltabuffers[lastLayerIdx];

		for (size_t i = 0; i < n->layers[lastLayerIdx].neurons.size(); i++) {
			delta[i] = (desired[i] - activations[i]) * activationDeriv(preactivations[i]);
		}
	}

	//ostatni vrstvy pocitaji s aktivacemi nasledujici vrstvy
	for (int layerIdx = static_cast<int>(lastLayerIdx) - 1; layerIdx >= static_cast<int>(layer); layerIdx--) {
		const std::vector<double>& preactivations = tlPreactivations[layerIdx];
		const std::vector<double>& nextDelta = deltabuffers[layerIdx + 1];
		const std::vector<NEURON>& nextLayerNeurons = n->layers[layerIdx + 1].neurons;
		const size_t layerSize = n->layers[layerIdx].neurons.size();
		std::vector<double>& delta = deltabuffers[layerIdx];

		for (size_t i = 0; i < layerSize; i++) {

			delta[i] = nextDelta[0] * nextLayerNeurons[0].weights[i + 1];
			for (size_t j = 1; j < nextLayerNeurons.size(); j++) {
				delta[i] += nextDelta[j] * nextLayerNeurons[j].weights[i + 1];
			}

			delta[i] *= activationDeriv(preactivations[i]);	//derivace dane aktivace nam da gradient v tomto bode
		}
	}
}

void adjustWeights(std::unique_ptr<NETWORK>& n, size_t layer, const std::vector<double>& adjustSourceActivation, const std::vector<double>& delta, double learningRate, double momentumRate) {		//upravi vahy dane vrstvy
	LAYER& l = n->layers[layer];			//dana vrstva
	const size_t layerSize = l.neurons.size();			//pocet neuronu v dane vrstve

	for (size_t i = 0; i < layerSize; i++) {		//pres neurony vrstvy
		NEURON& neuron = l.neurons[i];			//neuron z dane vrstvy

		const double weightedLearn = learningRate * delta[i];

		for (size_t j = 0; j < neuron.weights.size(); j++) {		//pres vahy vstupujici do neuronu

			const double term = momentumRate * (neuron.weights[j] - neuron.last_weights[j]); //momentum = setrvacnost z predchoziho kroku
			neuron.last_weights[j] = neuron.weights[j];
			neuron.weights[j] = neuron.weights[j] + term + weightedLearn * adjustSourceActivation[j];
		}

	}
}

void checkArgs(int argc, char** argv, int &inputs, int &outputs, int &samples, int &layers, std::vector<int> &npl, int &initW, std::vector<double>& wValues, int &showEqms, int &showEpochs, int &notShowWeights) {
	int i;
	for (i = 1; i < argc; i++) {

		char c = argv[i][0];
		char p = argv[i][1];

		if (c == '-') {
			if (p == 'o') {
				outputs = atoi(argv[++i]);
			}
			else if (p == 'i') {
				inputs = atoi(argv[++i]);
			}
			else if (p == 's') {
				shuffle = true;
			}
			else if (p == 'c') {
				classificationMode = true;
			}
			else if (p == 'f') {
				char sp = argv[i][2];
				if (sp == 'i') {
					inputFile = argv[++i];
				}
				else if (sp == 'm') {
					modelFile = argv[++i];
				}
			}
			else if (p == 'b') {
				methodBlitz = 1;
				char b = argv[i][2];
				if (b >= '1' && b <= '9')
					methodBlitz = b - '0';
				std::cout << "Blitz verze " << b << std::endl;
			}
			else if (p == 'q') {
				methodFreely = 1;
				char b = argv[i][2];
				if (b >= '1' && b <= '9')
					methodFreely = b - '0';
				std::cout << "Freely growing verze " << b << std::endl;
			}
			else if (p == 'l') {
				int j;
				layers = atoi(argv[++i]);
				std::vector<int> n(layers);

				for (j = 0; j < layers; j++) {
					n[j] = atoi(argv[++i]);
				}
				npl = n;
			}
			else if (p == 'w') {
				int j;
				int total = (inputs)+1;
				std::vector<int> n = npl;

				for (j = 0; j < layers; j++) {
					if (j == 0)
						total *= n[0];
					else
						total += (n[j - 1] + 1) * n[j];
				}
				initW = total;
				std::vector<double> w(total);
				wValues = w;
			}
			else if (p == 'r') {
				RATE = std::stod(std::string(&argv[i][3]));
				std::cout << "Mira uceni = " << RATE << std::endl;
			}
			else if (p == 'm') {
				MOMENTUM = std::stod(std::string(&argv[i][3]));
				std::cout << "Momentum = " << MOMENTUM << std::endl;
			}
			else if (p == 'x') {
				learnType = LearnType::None;
				std::cout << "Bez dalsiho algoritmu uceni" << std::endl;
			}
			else if (p == 'g') {

				char sp = argv[i][2];
				if (sp == 'n') {
					learnType = LearnType::Genetic_Algorithm_NoBackprop;
					std::cout << "Uceni genetickym algoritmem bez backpropagation" << std::endl;
				}
				else if (sp == 'm') {
					learnType = LearnType::Meta_Backprop;
					std::cout << "Uceni metaheuristikou s backpropagation" << std::endl;
				}
				else if (sp == 'b') {
					learnType = LearnType::Genetic_Algorithm_Backprop;
					std::cout << "Uceni genetickym algoritmem s backpropagation" << std::endl;
				}

				if (argc > i + 1) {

					std::string spec = argv[++i];
					std::istringstream iss(spec);

					std::string popStr, genStr;
					std::getline(iss, popStr, ',');
					std::getline(iss, genStr, ',');

					try {
						population_size_request = std::stoull(popStr);
						generations_count_request = std::stoull(genStr);

						std::cout << "Velikost populace = " << population_size_request << ", pocet generaci = " << generations_count_request << std::endl;
					}
					catch (...) {
						std::cout << "Neplatne zadana populace a generace, pouzivam vychozi" << std::endl;
					}
				}
			}
			else if (p == 'v') {
				validationMode = true;
				std::cout << "Rezim validace" << std::endl;
			}
		}
	}
}

void showUsage() {

	std::cout << "Pouziti:" << std::endl
		<< "\t./mlp -i <pocet> -o <pocet> -l <poc.vrstev> n1 n2 .. nN -fi <cesta> -fm <cesta>" << std::endl
		<< "" << std::endl
		<< "	-i <pocet> = pocet vstupnich priznaku" << std::endl
		<< "	-o <pocet> = pocet vystupu" << std::endl
		<< "	-l <poc.vrstev> n1 n2 .. nN = pocet vrstev a neuronu v nich, zahrnuje posledni vrstvu" << std::endl
		<< "	-fi <cesta> = cesta ke vstupnimu souboru" << std::endl
		<< "	-fm <cesta> = cesta k souboru s modelem" << std::endl
		<< "	-s = nahodne zamichani prvku pred ucenim" << std::endl
		<< "	-v = rezim validace" << std::endl
		<< "	-w = nacte vahy ze souboru pred ucenim, validaci nebo klasifikaci" << std::endl
		<< "	-b <mod> = spusti metodu Blitz s modem 1 nebo 2" << std::endl
		<< "	-q <mod> = spusti metodu Freely Growing s modem 1, 2 nebo 3" << std::endl
		<< "	-c = rezim klasifikace" << std::endl
		<< "	-x = zakaze dalsi uceni" << std::endl
		<< "	-r <cislo> = nastavi miru uceni na zadane cislo" << std::endl
		<< "	-m <cislo> = nastavi setrvacnost" << std::endl
		<< "	-gn <populace>,<generace> = spusti uceni diferencialni evoluci s danymi parametry" << std::endl
		<< "	-gb <populace>,<generace> = spusti uceni diferencialni evoluci s backpropem s danymi parametry" << std::endl
		<< "	-gm <populace>,<generace> = spusti uceni MetaBP s danymi parametry" << std::endl
		<< std::endl;
}
